package com.nisum.ecomproduct.controller;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nisum.ecomproduct.exception.ProductNotFoundException;
import com.nisum.ecomproduct.model.Product;
import com.nisum.ecomproduct.service.ProductService;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.WebMvcTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.MvcResult;

import java.util.Arrays;
import java.util.List;

import static org.hamcrest.Matchers.hasSize;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

@WebMvcTest(ProductController.class)
class ProductControllerTest {

    @Autowired
    MockMvc mockMvc;

    @MockBean
    ProductService productService;

    @Autowired
    ObjectMapper objectMapper;

    private String randomId = "someHexString";

    @Test
    void saveProduct() throws Exception {
        Product productRequestObject = getProductRequestObject();
        Product productResponseObject = getProductResponseObject();

        when(productService.saveProduct(any())).thenReturn(productResponseObject);
        MvcResult result = mockMvc.perform(post("/product/save")
                .content(objectMapper.writeValueAsString(productRequestObject))
                .contentType(MediaType.APPLICATION_JSON))
                .andExpect(status().isCreated())
                .andReturn();
        Product product = objectMapper.readValue(result.getResponse().getContentAsString(), Product.class);
        assertNotNull(product);
        assertNotNull(product.getId());
    }

    @Test
    void saveProducts() throws Exception {
        List<Product> productsRequest = getProductRequestObjectList();
        List<Product> productsResponse = getProductResponseObjectList();

        when(productService.saveProducts(any())).thenReturn(productsResponse);
        mockMvc.perform(post("/product/saveall")
                .content(objectMapper.writeValueAsString(productsRequest))
                .contentType(MediaType.APPLICATION_JSON))
                .andExpect(jsonPath("$", hasSize(3)))
                .andExpect(jsonPath("$[0].id", is(randomId)))
                .andExpect(jsonPath("$[1].id", is(randomId)))
                .andExpect(jsonPath("$[2].id", is(randomId)))
                .andExpect(status().isCreated())
                .andReturn();
    }

    @Test
    void getProductById_returnsProduct_200() throws Exception {
        when(productService.getProductById(randomId)).thenReturn(getProductResponseObject());
        mockMvc.perform(get("/product/id/" + randomId))
                .andExpect(status().isOk())
                .andExpect(jsonPath("$.id", is(randomId)))
                .andReturn();
    }

    @Test
    void getProductById_throws_404() throws Exception {
        when(productService.getProductById(randomId)).thenThrow(new ProductNotFoundException());
        mockMvc.perform(get("/product/id/" + randomId))
                .andExpect(status().isNotFound());
    }

    @Test
    void getRelatedProducts() {
    }

    @Test
    void updateProduct_returnsProduct_200() throws Exception {
        Product productRequestObject = getProductRequestObject();
        productRequestObject.setId(randomId);
        Product productResponseObject = getProductResponseObject();

        when(productService.updateProduct(any())).thenReturn(productResponseObject);
        MvcResult result = mockMvc.perform(put("/product/update")
                .content(objectMapper.writeValueAsString(productRequestObject))
                .contentType(MediaType.APPLICATION_JSON))
                .andExpect(status().isOk())
                .andReturn();
        Product product = objectMapper.readValue(result.getResponse().getContentAsString(), Product.class);
        assertNotNull(product);
        assertNotNull(product.getId());
    }

    @Test
    void updateProduct_throwsProductNotFoundException_404() throws Exception {
        Product productRequestObject = getProductRequestObject();
        productRequestObject.setId(randomId);

        when(productService.updateProduct(any())).thenThrow(new ProductNotFoundException());
        mockMvc.perform(put("/product/update")
                .content(objectMapper.writeValueAsString(productRequestObject))
                .contentType(MediaType.APPLICATION_JSON))
                .andExpect(status().isNotFound())
                .andReturn();
    }

    @Test
    void deleteProductById_204() throws Exception {
        doNothing().when(productService).deleteProductById(randomId);
        mockMvc.perform(delete("/product/" + randomId))
                .andExpect(status().isNoContent());
    }

    @Test
    void deleteProductById_throwsProductNotFoundException_404() throws Exception {
        doThrow(new ProductNotFoundException()).when(productService).deleteProductById(randomId);
        mockMvc.perform(delete("/product/" + randomId))
                .andExpect(status().isNotFound());
    }

    private Product getProductRequestObject() {
        Product product = new Product();
        product.setName("Dell Mouse");
        product.setDescription("Input device of the computer");
        product.setPrice(300D);
        return product;
    }

    private Product getProductResponseObject() {
        Product product = getProductRequestObject();
        product.setId(randomId);
        return product;
    }

    private List<Product> getProductRequestObjectList() {
        Product product1 = new Product();
        product1.setName("Bean bag with beans - L");
        product1.setDescription("Comfortable binge");
        product1.setPrice(800D);

        Product product2 = new Product();
        product2.setName("Bean bag with beans - XL");
        product2.setDescription("Comfortable binge");
        product2.setPrice(1000D);

        Product product3 = new Product();
        product3.setName("Bean bag with beans - XXL");
        product3.setDescription("IComfortable binge");
        product3.setPrice(1600D);

        return Arrays.asList(product1, product2, product3);
    }

    private List<Product> getProductResponseObjectList() {
        List<Product> products = getProductRequestObjectList();
        products.forEach(product -> product.setId(randomId));
        return products;
    }
}